#include <armadillo>

#include "LevenbergMarquardt.h"
#include "../message.h"

struct lib4neuro::LevenbergMarquardt::LevenbergMarquardtImpl {
    /**
 * Threshold for the successful ending of the optimization - deviation from minima
 */
    double tolerance;

    double tolerance_gradient;

    double tolerance_parameters;

    double LM_step_acceptance_threshold;

    double lambda_initial;

    double lambda_increase;

    double lambda_decrease;

    /**
     * Maximal number of iterations - optimization will stop after that, even if not converged
     */
    long long int maximum_niters;

    /**
     * Vector of minimum coordinates
     */
    std::vector<double> *optimal_parameters;

    /**
     * Returning Jacobian matrix of the residual function using the central differences method, i.e.
     * f'(x) = (f(x + delta) - f(x - delta)) / 2*delta
     *
     * @param h Step size
     * @return Jacobian matrix
     */
    arma::Mat<double>* get_Jacobian_matrix(lib4neuro::ErrorFunction& ef,
                                           arma::Mat<double>* J,
                                           double h=1e-3);
};

arma::Mat<double>* lib4neuro::LevenbergMarquardt::LevenbergMarquardtImpl::get_Jacobian_matrix(
        lib4neuro::ErrorFunction& ef,
        arma::Mat<double>* J,
        double h) {

    size_t n_parameters = ef.get_dimension();
    size_t n_data_points = ef.get_dataset()->get_n_elements();
    std::vector<std::pair<std::vector<double>, std::vector<double>>>* data = ef.get_dataset()->get_data();
    std::vector<double> grad(n_parameters);

    for(size_t i = 0; i < n_data_points; i++) {
        ef.calculate_residual_gradient(&data->at(i).first, &data->at(i).second, &grad, h);

        for(size_t j = 0; j < n_parameters; j++) {
            J->at(i, j) = grad.at(j);
        }
    }

    return J;
}

namespace lib4neuro {
    LevenbergMarquardt::LevenbergMarquardt(int max_iters,
                                           double tolerance,
                                           double tolerance_gradient,
                                           double tolerance_parameters,
                                           double LM_step_acceptance_threshold,
                                           double lambda_initial,
                                           double lambda_increase,
                                           double lambda_decrease) : p_impl(new LevenbergMarquardtImpl()) {

        this->p_impl->tolerance = tolerance;
        this->p_impl->tolerance_gradient = tolerance_gradient;
        this->p_impl->tolerance_parameters = tolerance_parameters;
        this->p_impl->LM_step_acceptance_threshold = LM_step_acceptance_threshold;
        this->p_impl->lambda_initial = lambda_initial;
        this->p_impl->lambda_increase = lambda_increase;
        this->p_impl->lambda_decrease = lambda_decrease;
        this->p_impl->maximum_niters = max_iters;
    }

    void LevenbergMarquardt::optimize(lib4neuro::ErrorFunction& ef,
                                      std::ofstream* ofs) {
        optimize(ef, LM_UPDATE_TYPE::MARQUARDT, ofs);
    }

    void LevenbergMarquardt::optimize(lib4neuro::ErrorFunction& ef,
                                      lib4neuro::LM_UPDATE_TYPE update_type,
                                      std::ofstream* ofs) {

        /* Copy data set max and min values, if it's normalized */
        if(ef.get_dataset()->is_normalized()) {
            ef.get_network_instance()->set_normalization_strategy_instance(
                    ef.get_dataset()->get_normalization_strategy());
        }

        COUT_INFO("Finding a solution via a Levenberg-Marquardt method with adaptive step-length..." << std::endl);
        if(ofs && ofs->is_open()) {
            *ofs << "Finding a solution via a Levenberg-Marquardt method with adaptive step-length..." << std::endl;
        }

        size_t n_parameters = ef.get_dimension();
        size_t n_data_points = ef.get_dataset()->get_n_elements();
        std::vector<double> *gradient_current = new std::vector<double>(n_parameters);
        std::fill(gradient_current->begin(), gradient_current->end(), 0.0);
        std::vector<double> *gradient_prev = new std::vector<double>(n_parameters);
        std::fill(gradient_prev->begin(), gradient_prev->end(), 0.0);
        std::vector<double> *params_current = ef.get_parameters();
        std::vector<double> *params_tmp = new std::vector<double>(n_parameters);
        arma::Mat<double> J(n_data_points, n_parameters);  // Jacobian matrix
        arma::Mat<double> H(n_data_points, n_parameters);  // Hessian matrix
        arma::Mat<double> H_new(n_data_points, n_parameters);

        std::vector<std::pair<std::vector<double>, std::vector<double>>>* data = ef.get_dataset()->get_data();

        double lambda = this->p_impl->lambda_initial;  // Dumping parameter
        double prev_err;
        double current_err;

        bool update_J = true;
        arma::Col<double> update;
        std::vector<double> d_prep(n_data_points);
        arma::Col<double>* d;

        //-------------------//
        // Solver iterations //
        //-------------------//
        size_t iter_counter = 0;
        do {
            if(update_J) {
                /* Get Jacobian matrix */
                this->p_impl->get_Jacobian_matrix(ef, &J);

                /* Get approximation of Hessian (H ~ J'*J) */
                H = J.t() * J;

                /* Evaluate the error before updating parameters */
                prev_err = ef.eval();
            }

            /* H_new = H + lambda*I */
            H_new = H + lambda * arma::eye<arma::Mat<double>>(n_parameters, n_parameters);

            /* Compute the residual vector */
            for(size_t i = 0; i < n_data_points; i++) {
                d_prep.at(i) = ef.calculate_single_residual(&data->at(i).first, &data->at(i).second);
            }
            d = new arma::Col<double> (d_prep);

            /* Compute the update vector */
            update = -H_new.i()*(J.t()*(*d));

            /* Compute the error after update of parameters */
            for(size_t i = 0; i < n_parameters; i++) {
                params_tmp->at(i) = update.at(i);
            }
            current_err = ef.eval(params_tmp);

            /* Check, if the parameter update improved the function */
            if(current_err < prev_err) {

                /* If the convergence threshold is achieved, finish the computation */
                if(current_err < this->p_impl->tolerance) {
                    break;
                }

                /* If the error is lowered after parameter update, accept the new parameters and lower the damping
                 * factor lambda */

                //TODO rewrite without division!
                lambda /= this->p_impl->lambda_decrease;

                for(size_t i = 0; i < n_parameters; i++) {
                    params_current->at(i) = update.at(i);
                }

                prev_err = current_err;
                update_J = true;

                COUT_DEBUG("Iteration: " << iter_counter << " Current error: " << current_err << std::endl);

            } else {
                /* If the error after parameters update is not lower, increase the damping factor lambda */
                update_J = false;
                lambda *= this->p_impl->lambda_increase;
            }

            COUT_DEBUG("Iteration: " << iter_counter << " Current error: " << current_err << std::endl);


        }while(iter_counter++ < this->p_impl->maximum_niters);

        /* Store the optimized parameters */
        this->p_impl->optimal_parameters = params_current;
    }

    std::vector<double>* LevenbergMarquardt::get_parameters() {
        return this->p_impl->optimal_parameters;
    }

    LevenbergMarquardt::~LevenbergMarquardt() = default;
}